#include "integrators.hpp"
#include "function.hpp"

// Integral calculation with Gauss-Chebyshev Quadrature.
// This is the part of NACRE project.
// NACRE -- Numerical Analysis / Computer REsearch.
//-------------------------------------------------------------------

// Copyleft (c) 2006 Alexey Beshenov <bav.272304@gmail.com>.

// NACRE is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation.

// NACRE is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details (the COPYING file).

//-----------------------------------------------------------------------------

void usage();

int main (int argc, char* argv[])
    {

    cout << "Integral calculation with Gauss-Chebyshev Quadrature" << endl;
    cout << "V.2006-04-17 [" << LIB_ID << "], copyleft (c) 2006 Alexey Beshenov" << endl;
    cout << COPY_NOTICE << endl << endl;

// number of arguments including keys, values, program name:
    int pn=9;

    if( argc != pn && argc != pn+2 ) { usage(); exit(EXIT_FAILURE); }

//-----------------------------------------------------------------------------

    int c, f, g, n; long double a, b;
    bool a_def=0, b_def=0, f_def=0, g_def=0, n_def=0;

    for (c=1; c<argc; c++) {
        if (!strcmp(argv[c], "-a")) {
            if (c<argc-1) { a = strtold(argv[c+1], NULL); a_def=1; }
            }
        if (!strcmp(argv[c], "-b")) {
            if (c<argc-1) { b = strtold(argv[c+1], NULL); b_def=1; }
            }
        if (!strcmp(argv[c], "-f")) {
            if (c<argc-1) { f = c+1; f_def=1; }
            }
        if (!strcmp(argv[c], "-g")) {
            if (c<argc-1) { g = c+1; g_def=1; }
            }
        if (!strcmp(argv[c], "-n")) {
            if (c<argc-1) { n = (int)strtold(argv[c+1], NULL); n_def=1; }
            }
        }

//-----------------------------------------------------------------------------

    if (n < 1) {
        cout << "ERROR! n must be an integer >= 1" << endl;
        exit(EXIT_FAILURE);
        }

    if (!a_def) { cerr << "ERROR! Missing parameter: -a" << endl; }
    if (!b_def) { cerr << "ERROR! Missing parameter: -b" << endl; }
    if (!f_def) { cerr << "ERROR! Missing parameter: -f" << endl; }
    if (!n_def) { cerr << "ERROR! Missing parameter: -n" << endl; }
    if (!(a_def&&b_def&&f_def&&n_def)) {
        usage(); exit(EXIT_FAILURE); cerr << endl;
        }

//-----------------------------------------------------------------------------

    if (!g_def) {                                 // function of real variable
        try { parser.Compile(addbrackets(argv[f])); }
        catch(TError error) { puts(error.error); exit(EXIT_FAILURE); }
        cout << "\\int_{" << a << "}^{" << b << "} [" << argv[f] << "]dx" << endl;
        cout.precision(OUTPUT_EPS); cout.setf(ios_base::fixed);
        cout << "~= " << gauss_chebyshev_calculate(n,a,b) << endl;
        }

    else {                                        // function of complex variable
        long double Re, Im;
        try { parser.Compile(addbrackets(argv[f])); }
        catch(TError error) { puts(error.error); exit(EXIT_FAILURE); }
        Re = gauss_chebyshev_calculate(n,a,b);
        try { parser.Compile(addbrackets(argv[g])); }
        catch(TError error) { puts(error.error); exit(EXIT_FAILURE); }
        Im = gauss_chebyshev_calculate(n,a,b);
        cout << "\\int_{" << a << "}^{" << b << "} [(" << argv[f] << ") + i*(" << argv[g] << ")]dx" << endl;
        ostringstream z;
        z.precision(OUTPUT_EPS); z.setf(ios_base::fixed);
        z << Re;
        if (fabs(Im) > EPS_LIM_T) { z << " + i*" << Im; }
        cout << "~= " << z.str() << endl;
        }

    PROC_TIME

        return 0;

    }


//-----------------------------------------------------------------------------

void usage()
    {
    cout << "Usage: -a <long double> -b <long double> -f \"<string>\" [-g \"<string>\"] -n <int>" << endl << endl;
    cout << "[z(x)]dx, z(x) = f(x) + i*g(x) if -g defined (function of complex variable)" << endl;
    cout << "[f(x)]dx if -g not defined (function of real variable)" << endl;
    cout << "Interval: [a, b]" << endl;
    cout << "n = 1, 2, 3, ..." << endl << endl;
    cout << builtin_functions;
    }
