// NACRE-integrators. Numerical integration and related operations.
// This is the part of NACRE project.
// NACRE -- Numerical Analysis / Computer REsearch.
//-------------------------------------------------------------------

// Copyleft (c) 2006 Alexey Beshenov <bav.272304@gmail.com>.

// Many Thanks: Dmitry Beshenov, Alexender Shargin.

// NACRE is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation.

// NACRE is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details (the COPYING file).

//-------------------------------------------------------------------

// Custom parameters listed below. Customize integrators by setting
// preprocessor macros.

// Order of default Newton-Cotes method
#define PPA_ALG 10

// Limitations
#define LIMIT_EPS_B true
#define EPS_LIM_B 1e-2
#define LIMIT_EPS_T true
#define EPS_LIM_T 1e-16
/* About 1e-18 at max on 32-bit systems (CHECK IT!)
We can't work with precision < EPS_LIM on processor arithmetic.
Decrease EPS_LIM_T on 64-bit machines. */
#define LIMIT_ITER true
#define ITER_LIM 1e6
/* It is STRONGLY recommended to limit iterations.
Use double check with both EPS_LIM's and ITER_LIM. */

// Signs after the floating point (output format).
#define OUTPUT_EPS 16                             // Usually -log10(EPS_LIM_T)

// Count and print working time
#define PRINT_PROC_TIME true

//-------------------------------------------------------------------

// ql-lib version
#define LIB_ID "NACRE-integrators-0.8.9"          // DO NOT change
// DO NOT remove copyright notice:
#define COPY_NOTICE "Program is distributed under the terms of the GNU GPL, see COPYING"

// Constants with precision 1e-40 form the "TAOCP" by D. E. Knuth
#define CNST_PI 3.1415926535897932384626433832795028841972
#define CNST_E 2.7182818284590452353602874713526624977572

//-------------------------------------------------------------------

#include <cmath>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

using namespace std;

//-------------------------------------------------------------------

// Coefficient matrix for polynomial:
typedef vector<long double, allocator<long double> > polynomial;

// Array of long double's:
typedef vector<long double, allocator<long double> > ld_array;

//-------------------------------------------------------------------

long double f (long double x);                    // Function to integrate (function.hpp)

//-------------------------------------------------------------------

// Newton-Cotes Quadrature
long double newton_cotes_coefficient (int n, int j)
// n = 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, j = n + 1
    {

    long double coefficient_table[10][11] = {
        1.0L,     1.0L,      0.0L,      0.0L,      0.0L,       0.0L,      0.0L,       0.0L,      0.0L,      0.0L,      0.0L,
        1.0L,     4.0L,      1.0L,      0.0L,      0.0L,       0.0L,      0.0L,       0.0L,      0.0L,      0.0L,      0.0L,
        1.0L,     3.0L,      3.0L,      1.0L,      0.0L,       0.0L,      0.0L,       0.0L,      0.0L,      0.0L,      0.0L,
        7.0L,     32.0L,     12.0L,     32.0L,     7.0L,       0.0L,      0.0L,       0.0L,      0.0L,      0.0L,      0.0L,
        19.0L,    75.0L,     50.0L,     50.0L,     75.0L,      19.0L,     0.0L,       0.0L,      0.0L,      0.0L,      0.0L,
        41.0L,    216.0L,    27.0L,     272.0L,    27.0L,      216.0L,    41.0L,      0.0L,      0.0L,      0.0L,      0.0L,
        751.0L,   3577.0L,   1323.0L,   2989.0L,   2989.0L,    1323.0L,   3577.0L,    751.0L,    0.0L,      0.0L,      0.0L,
        989.0L,   5888.0L,   -928.0L,   10496.0L,  -4540.0L,   10496.0L,  -928.0L,    5888.0L,   989.0L,    0.0L,      0.0L,
        2857.0L,  15741.0L,  1080.0L,   19344.0L,  5778.0L,    5778.0L,   19344.0L,   1080.0L,   15741.0L,  2857.0L,   0.0L,
        16067.0L, 106300.0L, -48525.0L, 272400.0L, -260550.0L, 427368.0L, -260550.0L, 272400.0L, -48525.0L, 106300.0L, 16067.0L
        };

// We can continue table above.

    return coefficient_table[n-1][j%(n+1)];

    }


// n = 1, 2, 3, 4, 5, 6, 7, 8, 9, 10
long double newton_cotes_calculate (long double a, long double b, int n, long double (*function) (long double x) = f)
    {

    long double sum = 0.0L;

    int j = 0;

    long double h = (b-a)/(long double)n;

    for (; j <= n; j++) {
        sum += function(a + (long double)j*h) * newton_cotes_coefficient (n, j);
        }

    long double P_n[10] = { 2.0L, 6.0L, 8.0L, 90.0L, 288.0L, 840.0L, 17280.0L, 28350.0L, 89600.0L, 598752.0L };

    return sum*(long double)n*h/P_n[n-1];

    }


//-------------------------------------------------------------------

// Composite Quadratures
long double composite_calculate (long double a, long double b, int n, int s, long double (*function) (long double x) = f)
    {
    int i = 1;
    long double h = (b - a)/(long double)s;
    long double sum = 0.0L;

    for (; i <= s; i++) {
        sum += newton_cotes_calculate (a+(long double)(i-1)*h, a+(long double)i*h, n, function);
        }

    return sum;
    }


//-------------------------------------------------------------------

// Chebyshev Quadrature
// n = 2, 3, 4, 5, 6, 7, 9
long double chebyshev_calculate (int n, long double a = -1.0L, long double b = 1.0L, long double (*function) (long double x) = f)
    {

    long double n_i[9][9] = {

// In this case "1.0L" means "no root(s)"
        1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.577350269189626L, 0.577350269189626L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.707106781186547L, 0.0L, 0.707106781186547L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.794654472316761L, -0.187592473979200L, 0.187592473979200L, 0.794654472316761L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.832497487082930L, -0.374541409440541L, 0.0L, 0.374541409440541L, 0.832497487082930L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.866246817876534L, -0.422518653761370L, -0.266635401516360L, 0.266635401516360L,  0.422518653761370L, 0.866246817876534L, 1.0L, 1.0L, 1.0L,

        -0.883861700758004L, -0.529656775285444L, -0.323911810519559L, 0.0L, 0.323911810519559L, 0.529656775285444L, 0.883861700758004L, 1.0L, 1.0L,

        1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L, 1.0L,

        -0.911589307722519L, -0.601018655464880L, -0.528761782967589L, -0.167906184214804L, 0.0L, 0.167906184214804L, 0.528761782967589L, 0.601018655464880L, 0.911589307722519L

        };

// We can improve table above by re-calculating roots of polynomials:

// (2)   x^2 - 1/3,
// (3)   x^3 - 0.5*x,
// (4)   x^4 - 2/3*x^2 + 1/45,
// (5)   x^5 - 5/6*x^3 + 7/72*x,
// (6)   x^6 - x^4 +0.2*x^2 - 1/105,
// (7)   x^7 - 7/6*x^5 + 119/360*x^3 - 149/6480*x,
// (9)   x^9 - 1.5*x^7 + 0.675*x^5 - 57/560*x^3 + 53/22400*x.

// We have calculated roots with precision 1.0e-15.

    int i = 1;

    long double sum = 0.0L;

    for (; i <= n; i++) {
        if (a!=-1.0L || b!=1.0L) {
// Using substitution to scale [a,b] to [-1,1].
            n_i[n-1][i-1] = 0.5L*((b-a)*n_i[n-1][i-1]+a+b);
            }
        sum += function(n_i[n-1][i-1]);

        }

    sum *= 2.0L/(long double)n;

    if (a!=-1.0L || b!=1.0L) {
        sum *= 0.5L*(b-a);
        }

    return sum;

    }


//-------------------------------------------------------------------

// Gauss-Chebyshev Quadrature
long double gauss_chebyshev_calculate (int n, long double a = -1.0L, long double b = 1.0L, long double (*function) (long double x) = f)
    {

    long double x_i, sum = 0.0L; int i;

    for (i=1; i<=n; i++) {
        x_i = cos(CNST_PI*((long double)i-0.5L)/(long double)n);

        if (a!=-1.0L || b!=1.0L) {
// Using substitution to scale [a,b] to [-1,1].
            sum += function(0.5L*((b-a)*x_i+a+b)) * sqrt(1.0L - pow(x_i, 2.0L));
            }
        else {
            sum += function(x_i) * sqrt(1.0L - pow(x_i, 2.0L));
            }

        }

    sum *= CNST_PI/(long double)n;

    if (a!=-1.0L || b!=1.0L) {
        sum *= 0.5L*(b-a);
        }

    return sum;

    }


//-------------------------------------------------------------------

// Gauss-Legendre Quadrature

// The following code was taken from NUMERICAL RECIPES IN C: THE ART OF SCIENTIFIC COMPUTING

// TODO: rewrite gauleg() AND gauss_legendre_calculate() WITH VECTORS!

void gauleg(long double a, long double b, long double x[], long double w[], int n)
    {
    int m,j,i;
    long double z1,z,xm,xl,pp,p3,p2,p1;
    m=(n+1)/2;
    xm=(b+a)/2.0;
    xl=(b-a)/2.0;
    long double eps = EPS_LIM_T*1e-1L;
    for (i=1;i<=m;i++) {
        z=cos(CNST_PI*((long double)i-0.25L)/((long double)n+0.5L));
        do {
            p1=1.0L;
            p2=0.0L;
            for (j=1;j<=n;j++) {
                p3=p2;
                p2=p1;
                p1=((2.0L*(long double)j-1.0L)*z*p2-((long double)j-1.0L)*p3)/(long double)j;
                }
            pp=n*(z*p1-p2)/(z*z-1.0L);
            z1=z;
            z=z1-p1/pp;
            } while (fabs(z-z1) > eps);
        x[i]=xm-xl*z;
        x[n+1-i]=xm+xl*z;
        w[i]=2.0L*xl/((1.0L-z*z)*pp*pp);
        w[n+1-i]=w[i];
        }
    }


long double gauss_legendre_calculate (long double a, long double b, int n, long double (*function) (long double x) = f)
    {

    long double *x = new long double [n+1];
    long double *w = new long double [n+1];

    gauleg(a, b, x, w, n);

    int i = 1;
    long double sum = 0;

    for (; i < n; i++) {

        sum += w[i]*function(x[i]);

        }

    return sum;

    }


//-------------------------------------------------------------------

// PPA (Prefered Precision Algorithm, as known as Romberg quadrature)
long double PPA_calculate (long double a, long double b, long double eps, int alg, int k=2, long double (*function) (long double x) = f)
    {

    long double before, after;

    before = composite_calculate(a,b,alg,(int)pow(2.0L,(long double)k), function);
    after = composite_calculate(a,b,alg,(int)pow(2.0L,(long double)(k+1)), function);

    while ( fabs ((before-after)*(b-a)) >= fabs (eps)) {

        before = composite_calculate(a,b,alg,(int)pow(2.0L,(long double)k), function);
        after = composite_calculate(a,b,alg,(int)pow(2.0L,(long double)(k+1)), function);
        k++;

        if ((int)pow(2.0L,(long double)(k+1)) > ITER_LIM && LIMIT_ITER) {
            cerr << "ERROR! Iterations limit exceeded (" << ITER_LIM << ")" << endl;
            cerr << "Precision probably got lost" << endl;
            break;
            }

        }

    return composite_calculate(a,b,alg,(int)pow(2.0L,(long double)(k+2)), function);

    }


//-------------------------------------------------------------------

#include "polynomial.hpp"

//-------------------------------------------------------------------

// Count and print working time
#define PROC_TIME clock_t proc_time = clock(); if (PRINT_PROC_TIME) { cout << endl << "Processor time: " << proc_time << " (" << proc_time/CLOCKS_PER_SEC <<" s)" << endl; }
